var Cc = Components.classes;
var Ci = Components.interfaces;

const Siemens = "Siemens PKCS#11 module";
const Charismathics = "Charismathics PKCS#11 module";
const OpenSC = "OpenSC PKCS#11 module";
const Bit4id = "Bit4id PKCS#11 module";

const FRIENDLY_FLAG = 0x1<<28

try{
	//Load ctypes if Firefox version is 4 or newer.
	Components.utils.import("resource://gre/modules/ctypes.jsm");
} catch(err) {}

function is_silent_module_install_posible() {
	try {
		return pkcs11 == null;
	} catch(err) {
		return true;
	}
}

function PKCS11Module( aName, aPath ) {
	this.name = aName;
	this.path = aPath;
	this.m_nStatus = 0;
}

PKCS11Module.prototype.__defineGetter__("status",
	function() {
		return this.m_nStatus;
	}
);

PKCS11Module.prototype.isOpenSCStdModule = function(path) {
	return (path == "opensc-pkcs11.so") || (path.indexOf("/opensc-pkcs11.so") > -1);
}

PKCS11Module.prototype.isRegistred = function( mod_name ) {
	var moddb = Cc["@mozilla.org/security/pkcs11moduledb;1"].getService(Ci.nsIPKCS11ModuleDB);
	var it = moddb.listModules();
	try {
		for( it.first(); ; it.next() ) {
			var mod = it.currentItem().QueryInterface(Ci.nsIPKCS11Module);
			if( mod.libName == null )
				continue;
			//If some of known modules is already registered trough Firefox GUI,
			//mod.libName will points to full path to PKCS#11 module, but we now use only library name.
			//So we must check if mod.libName ends with this.path to see if this is already registered.
			if(this.isOpenSCStdModule(this.path)) {
				if(this.isOpenSCStdModule(mod.libName)) {
					if( mod_name != null ) {
						mod_name.value = mod.name;
					}
					return true;
				}
			}
			else {
				var mod_path_end = mod.libName.toLowerCase().substring(mod.libName.length - this.path.length);
				if( mod_path_end == this.path.toLowerCase() ) {
					if( mod_name != null ) {
						mod_name.value = mod.name;
					}
					return true;
				}
			}
		}
	}catch(err) {}
	return false;
}

PKCS11Module.prototype.unregister = function(path)
{
	var tmp_name = {};
	var tmp_obj = new PKCS11Module("Temp", path);
	if( !tmp_obj.isRegistred(tmp_name) )
		return;
	if( !is_silent_module_install_posible() ) {
		//Gecko version < 1.9.1
		pkcs11.deleteModule(tmp_name.value);
	}
	else {
		//Gecko version >= 1.9.1
		var nsPKCS11 = Cc["@mozilla.org/security/pkcs11;1"].getService(Ci.nsIPKCS11);
		try {
			nsPKCS11.deleteModule(tmp_name.value);
		}
		catch( ex ) {}
	}
}

PKCS11Module.prototype.register = function() {
	if( !this.isRegistred() ) {
		//If module is onepin-opensc-pkcs11 remove opensc-pkcs11
		if( this.path.indexOf("onepin-opensc-pkcs11") > -1 ) {
			var temp_path = this.path.replace("onepin-opensc-pkcs11.so", "opensc-pkcs11.so");
			this.unregister(temp_path);
		}
		if( this.isOpenSCStdModule(this.path)) {
			var temp_path = this.path.replace("opensc-pkcs11.so", "onepin-opensc-pkcs11.so");
			this.unregister(temp_path);
		}
		if( !is_silent_module_install_posible() ) {
			//Gecko version < 1.9.1
			this.m_nStatus = pkcs11.addmodule( this.name, this.path, FRIENDLY_FLAG, 0 );
		}
		else {
			//Gecko version >= 1.9.1
			var nsPKCS11 = Cc["@mozilla.org/security/pkcs11;1"].getService(Ci.nsIPKCS11);
			try {
				nsPKCS11.addModule( this.name, this.path, FRIENDLY_FLAG, 0 );
				this.m_nStatus = 1;
			}
			catch( ex ) {
				this.m_nStatus = -100; // Registration failed with exception
			}
		}
	}
	else
		this.m_nStatus = -51; // Module exists
	return this.m_nStatus;
}

function PKCS11Setup() {
	this.isWindows = false;
	this.isLinux = false;
	this.isMac = false;

	this.sysdir = "";
	this.known_folders = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties);
	try{
		this.sysdir = (this.known_folders.get("SysD", Ci.nsIFile)).path;
	}
	catch( err ) {}

	var runtime = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULRuntime);
	this.osString = runtime.OS;
	this.processor = runtime.XPCOMABI.split("-")[0];
	//WINNT is Windows 2000 and above, WIN95 is Windows 95, Windows 98 or Windows Me
	if( this.osString.indexOf("WIN")!=-1 )
		this.isWindows = true;
	else if( this.osString.toLowerCase().indexOf("linux")!=-1 )
		this.isLinux = true;
	else if( this.osString == "Darwin" )
		this.isMac = true;
	//Settup for BSD and Solaris is the same as settup for Linux
	else if( this.osString.indexOf("BSD")!=-1 )
		this.isLinux = true;
	else if( this.osString == "SunOS" )
		this.isLinux = true;
}

PKCS11Setup.prototype.__defineGetter__("isOK",
	function() {
		return this.isWindows || this.isLinux || this.isMac;
	}
);

PKCS11Setup.prototype.is_pkcs11_lib = function(path) {
	//If there is no ctypes (Fireofx before 4) library can't be checked so just return true.
	if (typeof(ctypes) == "undefined")
		return true;
	
	try{
		//ctypes.open will search system library directories,
		//so there is no need to check all possible path for PKCS#11 libraries.
		var lib = ctypes.open(path);
		//PKCS#11 libraries must export C_GetFunctionList function.
		var func = lib.declare("C_GetFunctionList", ctypes.default_abi, ctypes.unsigned_long, ctypes.voidptr_t.ptr);
		lib.close();
		//There is no error, path points to real PKCS#11 library
		return true;
	} catch(e) {
		return false;
	}
}

PKCS11Setup.prototype.search_for_pkcs11_module = function() {
	var results = new Array();
	var arrCounter = 0;
	if( this.isWindows ) {
		if( this.processor == "x86_64" ) {
			if(this.is_pkcs11_lib("cmP1164.dll"))
				results[arrCounter++] = new PKCS11Module( Charismathics, "cmP1164.dll" );
		}
		else {
			if(this.is_pkcs11_lib("cmP11.dll"))
				results[arrCounter++] = new PKCS11Module( Charismathics, "cmP11.dll" );
		}
		if(this.is_pkcs11_lib("bit4ipki.dll"))
			results[arrCounter++] = new PKCS11Module( Bit4id, "bit4ipki.dll" );
		if(this.is_pkcs11_lib("siecap11.dll"))
			results[arrCounter++] = new PKCS11Module( Siemens, "siecap11.dll" );
	}
	else if( this.isLinux ) {
		if(this.is_pkcs11_lib("onepin-opensc-pkcs11.so"))
			results[arrCounter++] = new PKCS11Module( OpenSC, "onepin-opensc-pkcs11.so" );
		else if(this.is_pkcs11_lib("opensc-pkcs11.so"))
			results[arrCounter++] = new PKCS11Module( OpenSC, "opensc-pkcs11.so" );
		if(this.is_pkcs11_lib("libbit4ipki.so"))
			results[arrCounter++] = new PKCS11Module( Bit4id, "libbit4ipki.so" );
		if(this.is_pkcs11_lib("libsiecap11.so"))
			results[arrCounter++] = new PKCS11Module( Siemens, "libsiecap11.so" );
	}
	else if( this.isMac ) {
		if(this.is_pkcs11_lib("/Library/OpenSC/lib/onepin-opensc-pkcs11.so"))
			results[arrCounter++] = new PKCS11Module( OpenSC, "/Library/OpenSC/lib/onepin-opensc-pkcs11.so" );
		else if(this.is_pkcs11_lib("/Library/OpenSC/lib/opensc-pkcs11.so"))
			results[arrCounter++] = new PKCS11Module( OpenSC, "/Library/OpenSC/lib/opensc-pkcs11.so" );
		if(this.is_pkcs11_lib("/System/Library/bit4id/cryptoki/libbit4ipki.dylib"))
			results[arrCounter++] = new PKCS11Module( Bit4id, "/System/Library/bit4id/cryptoki/libbit4ipki.dylib" );
		if(this.is_pkcs11_lib("libsiecap11.dylib"))
			results[arrCounter++] = new PKCS11Module( Siemens, "libsiecap11.dylib" );
	}
	if( results.length > 0 )
		return results;
	return null;
}

// Return codes from https://developer.mozilla.org/en/PKCS11_Jar_Install
PKCS11Setup.prototype.setup = function() {
	var res = this.search_for_pkcs11_module();
	if( res != null )
		for( var i = 0; i < res.length; ++i )
			res[i].register();
	return res;
}

function message_box(message, title, icon) {
	//Possible values for icon: alert-icon, error-icon, message-icon, question-icon
	return openDialog("chrome://infonotary-mozilla-setup/content/messagebox.xul",
					  "IMS-msgBox", "chrome,centerscreen,modal,dialog",
					  message, title, icon);
}

function setup_pkcs11_module( gui ) {
	var strbundle = document.getElementById("inmozsetup-strings");
	var setup_res;
	var result = true;
	var obj = new PKCS11Setup();
	if( obj.isOK ) {
		setup_res = obj.setup();
	}
	else if( gui ) {
		var msg = strbundle.getFormattedString("osNotSupported", [ obj.osString ]);
		var title = strbundle.getString("inSetupPrefsTitle");
		message_box(title, msg, "error-icon");
	}

	if( setup_res == null )
		return false;
	
	var failed_modules = "";
	var installed_modules = "";
	for( var i = 0; i < setup_res.length; ++i ) {
		if( (setup_res[i].status < 0) && (setup_res[i].status != -51) )
			result = false;
		
		if( gui && (setup_res[i].status == -100) ) {
			failed_modules += setup_res[i].name + " (" + setup_res[i].path + ")\n";
			result = false;
		}
		
		if( gui && result ) {
			if( setup_res[i].status != -51 ) // Module exists, so it should not be added to installed_modules
				installed_modules += setup_res[i].name + " (" + setup_res[i].path + ")\n";
		}
	}
	if( failed_modules != "" ) {
		var msg = strbundle.getFormattedString("addModulesError", [failed_modules]);
		var title = strbundle.getString("inSetupPrefsTitle");
		message_box(msg, title, "alert-icon");
	}
	else if( gui ) {
		var strbundle = document.getElementById("inmozsetup-strings");
		var title = strbundle.getString("inSetupPrefsTitle");
		var msg = strbundle.getFormattedString("addModulesSuccess", [installed_modules]);
		message_box(msg, title, "message-icon");
	}
	return result;
}

function setup_is_needed( installed_modules_names ) {
	var obj = new PKCS11Setup();
	if( obj.isOK ) {
		var arr = obj.search_for_pkcs11_module();
		if( arr ) {
			var arrNames = new Array();
			var arrCounter = 0;
			var res = false;
			var name = {};
			for( var i = 0; i < arr.length; ++i ) {
				if( !arr[i].isRegistred(name) )
					res = true;
				arrNames[arrCounter++] = name.value;
			}
			if( installed_modules_names )
				installed_modules_names.value = arrNames.join(", ");
			return res;
		}
	}
	return false;
}
